'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _discord = require('discord.js');

var _discord2 = _interopRequireDefault(_discord);

var _eventemitter = require('eventemitter2');

var _eventemitter2 = _interopRequireDefault(_eventemitter);

var _lodash = require('lodash');

var _mongodb = require('mongodb');

var _hookManager = require('./hook-manager');

var _hookManager2 = _interopRequireDefault(_hookManager);

var _logger = require('./logger');

var _pluginManager = require('./plugin-manager');

var _pluginManager2 = _interopRequireDefault(_pluginManager);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * An application.
 */
class Application extends _eventemitter2.default {

  /**
   * Construct an application.
   */


  /**
   * The intervals.
   */


  /**
   * Indicate if the application is being started.
   */


  /**
   * The plugin manager.
   */


  /**
   * The Discord client.
   */

  /**
   * The application options.
   */
  constructor(options) {
    super();

    this.isStarting = false;
    this.isClosing = false;
    this.intervals = new Set();
    this.timeouts = new Set();
    this.options = options;
    this.logger = (0, _logger.createLogger)(options.logger);
    this.client = new _discord2.default.Client();
    this.pluginManager = new _pluginManager2.default(this);
    this.hookManager = new _hookManager2.default();

    this.client.on('warn', warning => this.logger.warn(warning));
    this.client.on('error', error => this.logger.error(error));

    this.client.on('ready', this.handleClientReady.bind(this));
    this.client.on('message', this.handleClientMessage.bind(this));
    this.client.on('messageUpdate', this.handleClientMessageUpdate.bind(this));
    this.client.on('disconnect', this.handleClientDisconnect.bind(this));
    this.client.on('reconnecting', this.handleClientReconnecting.bind(this));

    this.client.on('guildCreate', this.handleClientGuildCreate.bind(this));
    this.client.on('guildDelete', this.handleClientGuildDelete.bind(this));
    this.client.on('channelCreate', this.handleClientChannelCreate.bind(this));
    this.client.on('channelDelete', this.handleClientChannelDelete.bind(this));

    this.onAny((event, ...args) => {
      this.pluginManager.emit(event, ...args);
    });
  }

  /**
   * Start the application.
   */


  /**
   * The timeouts.
   */


  /**
   * Indicate if the application is being closed.
   */


  /**
   * The hook manager.
   */


  /**
   * The database connection.
   */


  /**
   * The logger instance.
   */
  async start() {
    if (!this.isStarting) {
      this.isStarting = true;

      this.logger.info('Connecting to database...');
      this.db = await _mongodb.MongoClient.connect(this.options.db.url, (0, _lodash.omit)(this.options.db, ['url']));

      this.logger.info('Loading plugins...');
      await this.pluginManager.loadPlugins();

      this.logger.info('Connecting to Discord servers...');
      await this.client.login(this.options.token);
    }
  }

  /**
   * Close the application.
   */
  async close() {
    if (!this.isClosing) {
      this.isClosing = true;

      this.timeouts.forEach(timeout => this.clearTimeout(timeout));
      this.intervals.forEach(interval => this.clearInterval(interval));

      this.logger.info('Disconnecting from Discord servers...');
      await this.client.destroy();

      this.logger.info('Clearing loaded plugins...');
      await this.pluginManager.clearPlugins();

      this.logger.info('Disconnecting from database...');
      await this.db.close();
    }
  }

  /**
   * Register a plugin.
   */
  register(plugin, options = {}) {
    this.pluginManager.registerPlugin(plugin, options);
  }

  /**
   * Extend the application with a new property.
   */
  decorate(name, value) {
    Object.defineProperty(Application.prototype, name, { value });
  }

  /**
   * Schedule a repeated execution.
   */
  setInterval(callback, delay, ...args) {
    const interval = setInterval(callback, delay, ...args);

    this.intervals.add(interval);
    return interval;
  }

  /**
   * Schedule a one-time execution.
   */
  setTimeout(callback, delay, ...args) {
    const timeout = setTimeout(callback, delay, ...args);

    this.timeouts.add(timeout);
    return timeout;
  }

  /**
   * Clear an interval.
   */
  clearInterval(interval) {
    clearInterval(interval);
    this.intervals.delete(interval);
  }

  /**
   * Clear a timeout.
   */
  clearTimeout(timeout) {
    clearTimeout(timeout);
    this.timeouts.delete(timeout);
  }

  /**
   * Called when the Discord client is ready.
   *
   * @private
   */
  handleClientReady() {
    this.logger.info('Application ready');
    this.emit('ready');
  }

  /**
   * Called when the Discord client disconnected.
   *
   * @private
   */
  handleClientDisconnect(event) {
    // Clear an unmanaged timeout preventing the application from being closed.
    clearTimeout(this.client.ws.connection.ratelimit.resetTimer);

    this.logger.info({ event }, 'Client disconnected');
    this.emit('disconnect', event);
  }

  /**
   * Called when the Discord client is reconnecting.
   *
   * @private
   */
  handleClientReconnecting() {
    this.logger.info('Client reconnecting...');
    this.emit('reconnecting');
  }

  /**
   * Called when the Discord client received a message.
   *
   * @private
   */
  async handleClientMessage(message) {
    if (message.author.bot) {
      return;
    }

    this.logger.trace({ message }, 'Message received');

    if (await this.hookManager.run(this, 'message', message, null)) {
      this.emit('message', message, null);
    }
  }

  /**
   * Called when the Discord client received a message update.
   *
   * @private
   */
  async handleClientMessageUpdate(oldMessage, newMessage) {
    if (newMessage.author.bot) {
      return;
    }

    this.logger.trace({ oldMessage, newMessage }, 'Message updated');

    if (await this.hookManager.run(this, 'message', newMessage, oldMessage)) {
      this.emit('message', newMessage, oldMessage);
    }
  }

  /**
   * Called when the Discord client created a guild.
   *
   * @private
   */
  async handleClientGuildCreate(guild) {
    this.logger.trace({ guild }, `Joined guild "${guild.name}"`);

    if (await this.hookManager.run(this, 'guildCreate', guild)) {
      this.emit('guildCreate', guild);
    }
  }

  /**
   * Called when the Discord client deleted a guild.
   *
   * @private
   */
  async handleClientGuildDelete(guild) {
    this.logger.trace({ guild }, `Left guild "${guild.name}"`);

    if (await this.hookManager.run(this, 'guildDelete', guild)) {
      this.emit('guildDelete', guild);
    }
  }

  /**
   * Called when the Discord client created a channel.
   *
   * @private
   */
  async handleClientChannelCreate(channel) {
    this.logger.trace({ channel }, `Channel "${channel.name}" created`);

    if (await this.hookManager.run(this, 'channelCreate', channel)) {
      this.emit('channelCreate', channel);
    }
  }

  /**
   * Called when the Discord client deleted a channel.
   *
   * @private
   */
  async handleClientChannelDelete(channel) {
    this.logger.trace({ channel }, `Channel "${channel.name}" removed`);

    if (await this.hookManager.run(this, 'channelDelete', channel)) {
      this.emit('channelDelete', channel);
    }
  }
}
exports.default = Application;